@@ -1,9 +1,6 @@
-use ammonia;
-use failure;
-
-use html5ever::rcdom::Node;
-use html5ever::rcdom::NodeData;
-use html5ever::rcdom::RcDom;
+use markup5ever_rcdom::Node;
+use markup5ever_rcdom::NodeData;
+use markup5ever_rcdom::RcDom;
 use html5ever::tendril::TendrilSink;
 use html5ever::tree_builder::TreeBuilderOpts;
 use html5ever::{parse_document, ParseOpts};
@@ -162,7 +159,7 @@ pub fn matrix_html_to_markup(s: &str) -> String {
         .to_string();
 
     markup_html(&sanitized_html)
-        .map(|x| x.trim().to_string())
+        .map(|x| x.trim_matches('\n').to_string())
         .unwrap_or_else(|_| markup_from_raw(&sanitized_html))
 }
 
@@ -237,6 +234,7 @@ fn convert_tag<'a>(t: &'a str) -> Option<(&'a str, &'a str)> {
         "strong",
         "b",
         "blockquote",
+        "li",
     ];
 
     if !allowed.contains(&t) {
@@ -250,16 +248,27 @@ fn convert_tag<'a>(t: &'a str) -> Option<(&'a str, &'a str)> {
         "br" => Some(("\n", "")),
         "code" => Some(("<tt>", "</tt>")),
         "strong" | "b" => Some(("<b>", "</b>")),
+        "li" => Some(("", "\n")),
         _ => None,
     }
 }
 
+fn match_tag(node: &Node, tags: &[&str]) -> bool {
+    match node.data {
+        NodeData::Element { name: ref n, .. } => {
+            let node_tag = n.local.to_string();
+            tags.contains(&node_tag.as_str())
+        }
+        _ => false,
+    }
+}
+
 fn convert_node(node: &Node, autolinks: bool) -> String {
     let mut output = String::new();
 
     match node.data {
         NodeData::Text { contents: ref c } => {
-            let escaped = &html_escape(&c.borrow().replace("\n", " "));
+            let escaped = &html_escape(&c.borrow().replace("\n", ""));
             if autolinks {
                 output.push_str(&markup_links(escaped));
             } else {
@@ -272,11 +281,38 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
             ..
         } => {
             let mut content = String::new();
-            for child in node.children.borrow().iter() {
-                content.push_str(&convert_node(child, true));
+            let tag = n.local.to_string();
+
+            match &tag[..] {
+                "ul" => {
+                    content.push_str("\n");
+                    for child in node.children.borrow().iter() {
+                        if match_tag(child, &["li", "ul", "ol"]) {
+                            content.push_str(&format!(" • {}", &convert_node(child, true)));
+                        }
+                    }
+                }
+                "ol" => {
+                    let mut counter = 1;
+                    content.push_str("\n");
+                    for child in node.children.borrow().iter() {
+                        if match_tag(child, &["li", "ul", "ol"]) {
+                            content.push_str(&format!(
+                                " {}. {}",
+                                counter.to_string(),
+                                &convert_node(child, true)
+                            ));
+                            counter += 1;
+                        }
+                    }
+                }
+                _ => {
+                    for child in node.children.borrow().iter() {
+                        content.push_str(&convert_node(child, true));
+                    }
+                }
             }
 
-            let tag = n.local.to_string();
             match &tag[..] {
                 "body" => {
                     output.push_str(&content);
@@ -331,7 +367,7 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
     output
 }
 
-pub fn markup_html(s: &str) -> Result<String, failure::Error> {
+pub fn markup_html(s: &str) -> Result<String, anyhow::Error> {
     let opts = ParseOpts {
         tree_builder: TreeBuilderOpts {
             drop_doctype: true,
